/**
 * Project: MentDB
 * Description: Mentalese Database Engine
 * Website: http://www.mentdb.org
 * Twitter: https://twitter.com/mentalese_db
 * LinkedIn: https://www.linkedin.com/company/mentdb
 * Facebook: https://www.facebook.com/mentdb
 * Author: Jimmitry Payet
 * Website: http://www.jpayet.re
 * Mail: contact@jpayet.re
 * Locality: Reunion Island (French)
 */

package re.jpayet.mentdb.core.db;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import re.jpayet.mentdb.core.basic.BasicData;
import re.jpayet.mentdb.core.basic.BasicIndex;
import re.jpayet.mentdb.core.basic.BasicPosition;
import re.jpayet.mentdb.core.basic.BasicRecordPosition;
import re.jpayet.mentdb.core.db.file.data.Block;
import re.jpayet.mentdb.core.db.file.data.DataFile;
import re.jpayet.mentdb.core.db.file.data.DataFilePool;
import re.jpayet.mentdb.core.db.file.data.UnusedFile;
import re.jpayet.mentdb.core.db.file.index.IndexFile;
import re.jpayet.mentdb.core.db.file.index.IndexFilePool;
import re.jpayet.mentdb.core.db.file.transaction.Transaction;

//Manage records
public class Record {

	public static LinkedHashMap<String, JSONObject> cache = new LinkedHashMap<String, JSONObject>();
	static public int RECORD_CACHE_SIZE = 0;
	
	public long dataFilePosition = -1;
	public String data = "";
	public int dataFileId = -1;
	
	//Constructor
	public Record(int dataFileId, long dataFilePosition, String data) {

		this.dataFileId = dataFileId;
		this.dataFilePosition = dataFilePosition;
		this.data = data;
		
	}
	
	public static JSONObject getNode(String key) throws IOException {
		
		//Try to get the node
		JSONObject result = cache.get(key);
		
		if (result==null) {
			
			//The node does not exist
			Record rec = Record.get_record(key);
			
			if (rec==null) {
				
				return null;
				
			} else {
			
				BasicData bd = new BasicData();
				bd.load(rec.data);
				result = bd.dataNode;
				
				//Save the node
				cache.put(key, result);
				
				//Delete the first if the size is too big
				if (cache.size()>RECORD_CACHE_SIZE) {
					
					Entry<String, JSONObject> e = cache.entrySet().iterator().next();
					
					cache.remove(e.getKey());
					
				}
				
				return result;
			
			}
			
		} else {
			
			//The node already exist
			return result;
			
		}
		
	}
	
	//Get a record
	public static Record get_record(String key) throws IOException {
		
		//Get the index file position
		int indexFileToWrite = IndexFile.indexFileId(key);
		
		//Get the position into the index file tab
		long indexFileTabPosition = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileToWrite).current_max_block_size());

		//Try to get the data
		Record r = IndexFile.existEntry(key, indexFileToWrite, indexFileTabPosition);

		return r;
		
	}
	
	//Update a record
	public static BasicRecordPosition update(String user, String key, String value) throws Exception {
		
		//Remove the record
		Record.remove(user, key);
		
		//Create the updated record
		BasicRecordPosition r = Record.add(user, key, value);
		
		return r;
		
	}
	
	//Add a new record
	public static BasicRecordPosition add(String user, String key, String value) throws Exception {
		
		//Get the index file position to write
		int indexFileId = IndexFile.indexFileId(key);
		
		//Get the position to write into the index file
		long indexFileTabIndex = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileId).current_max_block_size());
		
		//Try to get the record (data)
		Record record = IndexFile.existEntry(key, indexFileId, indexFileTabIndex);

		//Check if the key already exist
		if (record==null) {
			
			//Resize if > 75%
			if (IndexFilePool.get(indexFileId).isG75Percent()) {
			
				IndexFile.resize("mentdb", IndexFilePool.get(indexFileId));
				
				//Get the position to write into the index file
				indexFileTabIndex = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileId).current_max_block_size());
			
			}

			//Get the data file to write the new record (data)
			DataFile dataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(value));

			//Write and get the position in the data file
			long recordPosition = Record.write(user, key, dataFile, value);
			
			DataFile newCollisionDataFile = null;
			long newCollisionPosition = -1;

			//Check if the basic collision node does not exist
			if (!IndexFile.existCollisionNode(indexFileId, indexFileTabIndex)) {
				
				//Create the basic collision node and add the collision
				BasicIndex collisionNode = new BasicIndex();
				collisionNode.addCollision(key, dataFile.id, recordPosition);
				
				//Get the data file to write the collision node
				newCollisionDataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(collisionNode.collisionNode.toJSONString()));
				
				//Create the collision record
				newCollisionPosition = Record.write(user, key, newCollisionDataFile, collisionNode.collisionNode.toJSONString());
				
			} else {
				
				//Load the JSON collision
				BasicIndex collisionNode = new BasicIndex();
				String jsonCollisionNode = IndexFile.getCollisionNode(indexFileId, indexFileTabIndex);
				
				collisionNode.load(jsonCollisionNode);

				//Get old collision
				BasicPosition oldCollision = IndexFile.getCollisionCoord(indexFileId, indexFileTabIndex);
				
				//Delete the old collision
				Record.delete(key, user, DataFilePool.get(oldCollision.fileId) , oldCollision.position);
				
				//Add the new collision
				collisionNode.addCollision(key, dataFile.id, recordPosition);
				
				//Get the data file to write the collision
				newCollisionDataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(collisionNode.collisionNode.toJSONString()));
				
				//Create the collision record
				newCollisionPosition = Record.write(user, key, newCollisionDataFile, collisionNode.collisionNode.toJSONString());
				
			}
			
			//Write an entry in the index file
			IndexFilePool.get(indexFileId).writeEntry(user, indexFileId,
					indexFileTabIndex*12+8, 
					newCollisionDataFile.id, 
					newCollisionPosition);

			BasicRecordPosition result = new BasicRecordPosition(indexFileId, indexFileTabIndex*12+8, newCollisionDataFile.id, newCollisionPosition, dataFile.id, recordPosition);

			//Initialization
			RandomAccessFile logFiles = Transaction.logFiles.get(user);
			long logPosition = logFiles.length();
			
			//Move to the end position into the log file
			logFiles.seek(logPosition);
			
			//Write the begin of the block into the log file
			logFiles.writeUTF("0;A;"+indexFileId+";"+IndexFilePool.get(indexFileId).current_block_size());
			
			//Increment the file index
			IndexFilePool.get(indexFileId).increment();
			
			return result;
			
		} else return null;
		
	}
	
	//Recreate a new record
	public static void recreate(String user, String key, String value) throws Exception {
				
		//Get the index file position to write
		int indexFileId = IndexFile.indexFileId(key);
			
		//Get the position to write into the index file
		long indexFileTabIndex = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileId).current_max_block_size());

		//Try to get the record (data)
		Record record = IndexFile.existEntry(key, indexFileId, indexFileTabIndex);

		//Check if the key already exist
		if (record==null) {

			//Get the data file to write the new record (data)
			DataFile dataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(value));

			//Write and get the position in the data file
			long recordPosition = Record.write(user, key, dataFile, value);
			
			DataFile newCollisionDataFile = null;
			long newCollisionPosition = -1;

			//Check if the basic collision node does not exist
			if (!IndexFile.existCollisionNode(indexFileId, indexFileTabIndex)) {

				//Create the basic collision node and add the collision
				BasicIndex collisionNode = new BasicIndex();
				collisionNode.addCollision(key, dataFile.id, recordPosition);

				//Get the data file to write the collision node
				newCollisionDataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(collisionNode.collisionNode.toJSONString()));
				
				//Create the collision record
				newCollisionPosition = Record.write(user, key, newCollisionDataFile, collisionNode.collisionNode.toJSONString());
				
			} else {

				//Load the JSON collision
				BasicIndex collisionNode = new BasicIndex();
				collisionNode.load(IndexFile.getCollisionNode(indexFileId, indexFileTabIndex));
				
				//Get old collision
				BasicPosition oldCollision = IndexFile.getCollisionCoord(indexFileId, indexFileTabIndex);
				
				//Delete the old collision
				Record.delete(key, user, DataFilePool.get(oldCollision.fileId) , oldCollision.position);
				
				//Add the new collision
				collisionNode.addCollision(key, dataFile.id, recordPosition);

				//Get the data file to write the collision
				newCollisionDataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(collisionNode.collisionNode.toJSONString()));
				
				//Create the collision record
				newCollisionPosition = Record.write(user, key, newCollisionDataFile, collisionNode.collisionNode.toJSONString());
				
			}
			
			//Write an entry in the index file
			IndexFilePool.get(indexFileId).writeEntry(user, indexFileId,
					indexFileTabIndex*12+8, 
					newCollisionDataFile.id, 
					newCollisionPosition);
			
		};
		
	}
	
	//Remove a record
	public static BasicRecordPosition remove(String user, String key) throws Exception {
		
		//Remove the key from the cache
		cache.remove(key);
				
		//Get the index file position
		int indexFileId = IndexFile.indexFileId(key);
		
		//Get the position into the index file tab
		long indexFileTabIndex = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileId).current_max_block_size());

		//Try to get the record
		Record record = IndexFile.existEntry(key, indexFileId, indexFileTabIndex);

		if (record!=null) {

			//Remove the data node
			Record.delete(key, user, DataFilePool.get(record.dataFileId), record.dataFilePosition);
			
			//Load the JSON collision
			BasicIndex biNode = new BasicIndex();
			biNode.load(IndexFile.getCollisionNode(indexFileId, indexFileTabIndex));
			
			//Get old collision
			BasicPosition oldCollision = IndexFile.getCollisionCoord(indexFileId, indexFileTabIndex);
			
			//Delete the old collision record
			Record.delete(key, user, DataFilePool.get(oldCollision.fileId) , oldCollision.position);
			
			//Delete the collision
			biNode.deleteCollision(key);
			
			BasicRecordPosition result = null;
			
			if (biNode.countCollision()!=0) {

				//Get the data file to write the collision node
				DataFile newCollisionDataFile = DataFilePool.getDataFileToWrite(Record.nbBlockToWrite(biNode.collisionNode.toJSONString()));
				
				//Create the collision record
				long newCollisionPosition = Record.write(user, key, newCollisionDataFile, biNode.collisionNode.toJSONString());
				
				//Write an entry in the index file
				IndexFilePool.get(indexFileId).writeEntry(user, indexFileId,
						indexFileTabIndex*12+8, 
						newCollisionDataFile.id, 
						newCollisionPosition);
				
				result = new BasicRecordPosition(indexFileId, indexFileTabIndex*12+8, newCollisionDataFile.id, newCollisionPosition, record.dataFileId, record.dataFilePosition);
				
			} else {
				
				//Clear the index entry
				IndexFilePool.get(indexFileId).writeEntry(user, indexFileId,
						indexFileTabIndex*12+8, 
						-1, 
						-1);
				
				result = new BasicRecordPosition(indexFileId, indexFileTabIndex*12+8, -1, -1, record.dataFileId, record.dataFilePosition);
				
			}
			
			//Initialization
			RandomAccessFile logFiles = Transaction.logFiles.get(user);
			long logPosition = logFiles.length();
			
			//Move to the end position into the log file
			logFiles.seek(logPosition);
			
			//Write the begin of the block into the log file
			logFiles.writeUTF("0;A;"+indexFileId+";"+IndexFilePool.get(indexFileId).current_block_size());
			
			//Increment the file index
			IndexFilePool.get(indexFileId).decrement();

			return result;
			
		} else return null;
		
	}
	
	//Get a record details
	@SuppressWarnings("unchecked")
	public static JSONObject getDetail(String key) throws Exception {
				
		//Initialization
		JSONObject coord = new JSONObject();
		
		//Get the index file position
		int indexFileToWrite = IndexFile.indexFileId(key);
		coord.put("key", key);
		coord.put("indexFileToWrite", indexFileToWrite);
		
		//Get the position into the index file tab
		long indexFileTabPosition = IndexFile.indexFileTabIndex(key, IndexFilePool.get(indexFileToWrite).current_max_block_size());
		coord.put("indexFileTabPosition", indexFileTabPosition);
		
		Record rec = IndexFile.existEntry(key, indexFileToWrite, indexFileTabPosition);
		
		if (rec == null) {
			
			throw new Exception("Sorry, the key "+key+" does not exist.");
			
		}
		
		coord.put("dataFileId", rec.dataFileId);
		coord.put("dataFilePosition", rec.dataFilePosition);
		coord.put("data", rec.data);
		coord.put("size", rec.data.length());
		
		return coord;
		
	}
	
	//Delete a record
	public static void delete(String key, String user, DataFile dataFile, long position) throws IOException {
		
		//Get the block
		Block block = Block.read(dataFile, position);
		
		//Check if there is another block and delete it
		if (block.nextBlockAddress!=-1) {
			
			delete(key, user, dataFile, block.nextBlockAddress);
			
		}
		
		//Delete the block
		Block.do_delete(key, user, dataFile, position);
		
	}
	
	//Count the number of block to write
	public static int nbBlockToWrite(String recValue) throws IOException {
		
	    return (int) Math.ceil(recValue.length()/DataFile.BLOCK_SIZE)+1;
	    
	}
	
	//Write a record
	public static long write(String user, String key, DataFile dataFile, String recValue) throws IOException {
		
		return writeB64(user, key, dataFile, Block.encodeb64(recValue));
		
	}
	
	//Write a record
	public static long writeB64(String user, String key, DataFile dataFile, String recValue) throws IOException {
		
		//Initialization
		List<Block> blocks = new ArrayList<Block>();
		List<Long> blocksPosition = new ArrayList<Long>();
		
		//Create all blocks for the record
	    for (int i = 0; i < recValue.length(); i += DataFile.BLOCK_SIZE) {
	    	
	    	//Create the block
	    	
	    	Block block = new Block(recValue.substring(i, Math.min(recValue.length(), i + DataFile.BLOCK_SIZE)));
	    	
	    	//Save the block
	    	blocks.add(block);
	    	
	    	//Get and save a position
	    	blocksPosition.add(UnusedFile.getPosition(user, dataFile));
	    	
	    }
	    
	    //Write all blocks in the record
	    for (int i = 0; i < blocks.size(); i ++) {
	    	
	    	long position = blocksPosition.get(i);
	    	
	    	//Check if this is the last block
	    	if (i+1 == blocks.size()) {
	    		
	    		Block.do_write(user, key, dataFile, position, blocks.get(i), -1);
	    		
	    	} else {

		    	long positionP1 = blocksPosition.get(i+1);

	    		Block.do_write(user, key, dataFile, position, blocks.get(i), positionP1);
	    		
	    	}
	    	
	    }
	    
	    return blocksPosition.get(0);
		
	}
	
	//Get the data into the string format
	public static String getData(String key, DataFile dataFile, long position) throws IOException {
		
		//Initialization
		
		Block block = Block.read(dataFile, position);
		
		StringBuilder result = new StringBuilder();
		result.append(block.getData());
		
		if (block.nextBlockAddress!=-1) {
			
			result.append(getData(key, dataFile, block.nextBlockAddress));
			
		}
		
	    return result.toString();
		
	}
	
	//Read a record
	@SuppressWarnings("unchecked")
	public static String read(DataFile dataFile, long nb, long position, JSONArray array) throws IOException {
		
		//Initialization
		Block block = Block.read(dataFile, position);
		String recString = block.dataStr.substring(0, block.dataStrSize);
		
		JSONObject json = new JSONObject();
		json.put("blockId", nb);
		json.put("position", position);
		BasicData bd = new BasicData();
		bd.load(block.toString());
		json.put("block", bd.dataNode);
		array.add(json);
		
		
		if (block.nextBlockAddress!=-1) {
			
			recString += read(dataFile, nb+1, block.nextBlockAddress, array);
			
		}
		
		return recString;
		
	}

}
