/**
 * Project: MentDB
 * Description: Mentalese Database Engine
 * Website: http://www.mentdb.org
 * Twitter: https://twitter.com/mentalese_db
 * LinkedIn: https://www.linkedin.com/company/mentdb
 * Facebook: https://www.facebook.com/mentdb
 * Author: Jimmitry Payet
 * Website: http://www.jpayet.re
 * Mail: contact@jpayet.re
 * Locality: Reunion Island (French)
 */

package re.jpayet.mentdb.core.script;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Date;
import java.util.HashMap;
import java.util.Vector;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eclipse.jetty.util.URIUtil;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import re.jpayet.mentdb.core.basic.BasicData;
import re.jpayet.mentdb.core.command.Command;
import re.jpayet.mentdb.core.command.MQLValue;
import re.jpayet.mentdb.core.db.Record;
import re.jpayet.mentdb.core.db.file.json.JsonFormatter;
import re.jpayet.mentdb.core.db.file.transaction.Transaction;
import re.jpayet.mentdb.core.fx.AtomFx;
import re.jpayet.mentdb.core.fx.DateFx;
import re.jpayet.mentdb.core.fx.StringFx;
import re.jpayet.mentdb.core.session.SessionThread;
import re.jpayet.mentdb.core.session.env.EnvManager;
import re.jpayet.mentdb.core.user.GroupManager;
import re.jpayet.mentdb.core.user.UserManager;
import re.jpayet.mentdb.server.Start;
import re.jpayet.mentdb.tools.Misc;

//The script class
public class ScriptManager {
	
	//Script accelerator
	public static HashMap<String, JSONObject> scriptAccelerators = new HashMap<String, JSONObject>();
	
	//The constructor
	public ScriptManager() {
		
	}
	
	//This function is used by the JSP page for execute a specific container
	@SuppressWarnings("unchecked")
	public static String execute(HttpServletRequest request, HttpServletResponse response) throws Exception {
		
		//Initialization
		Date d1 = new Date();
		String result = "";
		boolean closeAfter = false;
		
		response.setContentType("application/json");
		response.setCharacterEncoding("UTF-8");

		//Load the authentication
		String token = request.getParameter("token");
		String script = request.getParameter("script");
		String user = "";
		
		//Try to execute a mental script
		try {
			
			//Return a message if the user is null or empty
			if (token==null || token.equals("")) {
				
				throw new Exception("Sorry, the token cannot be null or empty.");

			}
			
			//Return a message if the script is null or empty
			if (script==null || script.equals("")) {

				throw new Exception("Sorry, the script cannot be null or empty.");

			}
			
			//Get the user/password
			user = UserManager.get_user_from_token(token);
			
			if (Transaction.logFiles.containsKey(user)) {
				
				throw new Exception("Sorry, the user '"+user+"' is already connected.");
				
			}
			
			if (!GroupManager.isGrantedUser(user, "api-rest", user)) {
				
				throw new Exception("Sorry, '"+user+"' is not in 'api-rest' group."); 
				
			}
			
			if (!ScriptManager.isGrantedToUser("mentdb", user, script)) {
				
				throw new Exception("Sorry, the script '"+script+"' is not granted to the user '"+user+"'."); 
				
			}
			
			String mentalScript = "execute \""+script+"\"";
			SessionThread thread = new SessionThread(null);
			thread.user = user;
			
			//Create the data/transaction/user folder if does not exist
			if (!(new File("data"+File.separator+"transaction"+File.separator+user).exists())) {

				(new File("data"+File.separator+"transaction"+File.separator+user)).mkdir();

			}
			
			String allArguments = ScriptManager.get(script).get("vars")+"";
			
			int nbVars = 0;
			if (!StringFx.lrtrim(allArguments).equals("")) {
				nbVars = Integer.parseInt(AtomFx.size(allArguments, ","));
			}
			
			//Set all variables
			for(int i=1;i<=nbVars;i++) {
				
				String parameterName = StringFx.lrtrim(AtomFx.get(allArguments, ""+i, ","));
				String param = parameterName.substring(1, parameterName.length()-1);
				
				if (!request.getParameterMap().containsKey(param)) {

					throw new Exception("Sorry, the parameter '"+param+"' is required.");
					
				}

				mentalScript += " \""+parameterName+"\" \""+request.getParameter(param).replace("\"", "\\\"")+"\"";
				
			}
			
			closeAfter = true;
			
			//Open the log file
			Transaction.logFiles.put(user, new RandomAccessFile("data"+File.separator+"transaction"+File.separator+user+File.separator+"rollback.log", "rw"));
			
			result = Command.executeAllCommands(true, thread, Misc.splitCommand(mentalScript), thread.env);
			
			JSONObject valid = new JSONObject();
			valid.put("MentDB", Start.version);
			valid.put("status", "OK");
			Date d2 = new Date();
			long seconds = (d2.getTime()-d1.getTime())/1000;
			valid.put("time", seconds+"s");
			valid.put("user", user);
			valid.put("timestamp", DateFx.current_timestamp());
			JSONParser jsonParser = new JSONParser();
			
			try {
				
				JSONObject r = (JSONObject) jsonParser.parse(result);
				valid.put("result", r);
				
			} catch (Exception a) {
				
				try {
					
					JSONArray r = (JSONArray) jsonParser.parse(result);
					valid.put("result", r);
					
				} catch (Exception b) {
					valid.put("result", result);
				}
				
			}
			
			return JsonFormatter.format(valid.toJSONString());

		} catch (Exception e) {
			
			JSONObject error = new JSONObject();
			error.put("MentDB", Start.version);
			error.put("status", "KO");
			error.put("token", token);
			error.put("errorMsg", ""+e.getMessage());
			error.put("timestamp", DateFx.current_timestamp());
			
			return JsonFormatter.format(error.toJSONString());

		} finally {
			
			if (!user.equals("") && closeAfter) {
			
				try {
					
					Transaction.rollback(user);
					Misc.deleteFile("data"+File.separator+"transaction"+File.separator+user);
					Transaction.logFiles.remove(user);
					
				} catch (IOException e) {
					//e.printStackTrace();
				}
				
			}
			
		}
		
	}

	//Generate an URL to execute a mental script
	public static String generateUrl(String scriptName, String user) throws Exception {
		
		String result = "http://localhost:"+Start.WEB_SERVER_PORT+"/api-rest/?token="+UserManager.get_token_from_user(user)+"&script="+URIUtil.encodePath(scriptName);
		
		String allArguments = ScriptManager.get(scriptName).get("vars")+"";
		
		int nbVars = 0;
		if (!StringFx.lrtrim(allArguments).equals("")) {
			nbVars = Integer.parseInt(AtomFx.size(allArguments, ","));
		}
		
		//Set all variables
		for(int i=1;i<=nbVars;i++) {
			
			String parameterName = StringFx.lrtrim(AtomFx.get(allArguments, ""+i, ","));
			String param = parameterName.substring(1, parameterName.length()-1);

			result += "&"+URIUtil.encodePath(param)+"=";
			
		}
		
		return result;
		
	}

	//Generate execute script
	public static String generateExecute(String scriptName) throws Exception {
		
		String result = "execute \""+scriptName.replace("\"", "\\\"")+"\"";
		
		String allArguments = ScriptManager.get(scriptName).get("vars")+"";
		
		int nbVars = 0;
		if (!StringFx.lrtrim(allArguments).equals("")) {
			nbVars = Integer.parseInt(AtomFx.size(allArguments, ","));
		}
		
		//Set all variables
		for(int i=1;i<=nbVars;i++) {
			
			String parameterName = StringFx.lrtrim(AtomFx.get(allArguments, ""+i, ","));

			result += "\n\t\""+parameterName+"\" \"\"";
			
		}
		
		return result+"\n;";
		
	}

	//Generate include script
	public static String generateInclude(String scriptName) throws Exception {
		
		String result = "include \""+scriptName.replace("\"", "\\\"")+"\"";
		
		String allArguments = ScriptManager.get(scriptName).get("vars")+"";
		
		int nbVars = 0;
		if (!StringFx.lrtrim(allArguments).equals("")) {
			nbVars = Integer.parseInt(AtomFx.size(allArguments, ","));
		}
		
		//Set all variables
		for(int i=1;i<=nbVars;i++) {
			
			String parameterName = StringFx.lrtrim(AtomFx.get(allArguments, ""+i, ","));

			result += "\n\t\""+parameterName+"\" \"\"";
			
		}
		
		return result+"\n;";
		
	}
	
	//Initialize the script manager
	public static void init(String user) throws Exception {

		JSONObject dataNode = new JSONObject();
		
		Record.add(user, "SCRIPT[]", dataNode.toJSONString());

	}

	//Check if a script already exist
	public static boolean exist(String scriptName) throws Exception {
		
		//Get the SCRIPT object
		JSONObject bd = Record.getNode("SCRIPT[]");
		
		scriptName = scriptName.replace("[", "").replace("]", "");
		
		return bd.containsKey(scriptName);
		
	}

	//Get the MQL script
	public static JSONObject get(String scriptName) throws Exception {
		
		//Generate an error if the script does not exist
		if (!exist(scriptName)) {

			throw new Exception("Sorry, the script "+scriptName+" does not exist.");

		}
		
		return Record.getNode("script["+scriptName+"]");
		
	}

	//Execute a script
	public static String execute(SessionThread session, Vector<MQLValue> inputVector, EnvManager env) throws Exception {
		

		//Initialization
		String scriptName = inputVector.get(0).value;
		
		//Check if a script is granted to a user
		if (!session.user.equals("admin") && !session.user.equals("ai") && !session.user.equals("mentdb") && !ScriptManager.isGrantedToUser(session.user, session.user, scriptName)) {

			throw new Exception("Sorry, the script '"+scriptName+"' is not granted to the user '"+session.user+"'.");

		}
		
		int nbParamDiviseBy2 = (inputVector.size()-1)/2;
		if ((nbParamDiviseBy2*2)!=(inputVector.size()-1)) {

			throw new Exception("Sorry, the script does not has a valid number of parameters.");

		}
		
		//Get script object
		JSONObject script = null;
		if (scriptAccelerators.containsKey(scriptName)) {
			script = scriptAccelerators.get(scriptName);
		} else {
			script = ScriptManager.get(scriptName);
		}
		String vars = script.get("vars")+"";
		String mql = script.get("mql")+"";
		String mem = script.get("mem")+"";
		
		if (mem.equals("1")) {
			scriptAccelerators.put(scriptName, script);
		}
		
		//Set all parameters into the variable environment
		int posVar = 1;
		for(int i=1;i<inputVector.size();i=i+2) {

			//Get the variable and the value
			String varName = inputVector.get(i).value;
			String varRequired = Misc.atom(vars, posVar, ",");
			if (!varName.equals(varRequired)) {
				throw new Exception("Sorry, the variable "+varRequired+" is required in position "+posVar+" (found "+varName+").");
			}
			
			String val = inputVector.get(i+1).value;
			
			env.set(varName, val);
			
			posVar++;
		}
		
		//Execute MQL
		return Command.executeAllCommands(false, session, Misc.splitCommand(mql), env);
		
	}
	
	//Check if it is a variable name
	public static boolean is_var(String varName) {
		
		//Test
		if (varName.startsWith("[") && varName.endsWith("]") && varName.length()>2) {

			//Return the boolean
			if (varName.substring(1, varName.length()-1).indexOf("[")>-1 || varName.substring(1, varName.length()-1).indexOf("]")>-1) return false;
			else return true;

		} else {

			//Is not a valid string
			return false;

		}
	
	}
	
	//Generate update
	public static String generateUpdate(String scriptName) throws Exception {
		
		//Initialization
		String result = "script update \""+scriptName.replace("\"", "\\\"")+"\" ";
		JSONObject script = ScriptManager.get(scriptName);
		result += script.get("mem")+" ";
		result += "\""+(script.get("vars")+"").replace("\"", "\\\"")+"\" ";
		result += "{"+(script.get("mql")+"")+"};";
		
		return result;
		
	}
	
	//Show all scripts
	public static JSONObject show() throws Exception {
		
		return Record.getNode("SCRIPT[]");
		
	}
	
	//Unload accelerator
	public static void unload_accelerators() throws Exception {
		
		ScriptManager.scriptAccelerators = new HashMap<String, JSONObject>();
		
	}

	//Add a script
	@SuppressWarnings("unchecked")
	public static void add(String user, String scriptName, String varNames, String mql, String inMemory) throws Exception {
		
		//Generate an error if the script already exist
		if (exist(scriptName)) {

			throw new Exception("Sorry, the script "+scriptName+" already exist.");

		}
		
		int nbVars = 0;
		if (!StringFx.lrtrim(varNames).equals("")) {
			nbVars = Integer.parseInt(AtomFx.size(varNames, ","));
		}
		
		//Check the variable name format
		for(int i=1;i<=nbVars;i++) {
			
			String curVarName = Misc.lrtrim(Misc.atom(varNames, i, ","));
			if (!is_var(curVarName)) {
				throw new Exception("Sorry, "+curVarName+" is not a valid var name (ex: {var}).");
			}
		}
		
		scriptName = scriptName.replace("[", "").replace("]", "");
		
		//Get the script object
		JSONObject bd = Record.getNode("SCRIPT[]");
		
		JSONObject obj = new JSONObject();
		obj.put("k", scriptName);
		obj.put("vars", varNames);
		obj.put("mql", mql);
		obj.put("mem", inMemory);
		
		Record.add(user, "script["+scriptName+"]", obj.toJSONString());
		
		bd.put(scriptName, 0);

		Record.update(user, "SCRIPT[]", bd.toJSONString());
		
	}

	//Update a script
	@SuppressWarnings("unchecked")
	public static void update(String user, String scriptName, String varNames, String mql, String inMemory) throws Exception {
		
		//Generate an error if the script does not exist
		if (!exist(scriptName)) {

			throw new Exception("Sorry, the script "+scriptName+" does not exist.");

		}
		
		int nbVars = 0;
		if (!StringFx.lrtrim(varNames).equals("")) {
			nbVars = Integer.parseInt(AtomFx.size(varNames, ","));
		}
		
		//Check the variable name format
		for(int i=1;i<=nbVars;i++) {
			
			String curVarName = Misc.atom(varNames, i, ",");
			if (!is_var(curVarName)) {
				throw new Exception("Sorry, "+curVarName+" is not a valid var name (ex: {var}).");
			}
		}
		
		scriptName = scriptName.replace("[", "").replace("]", "");
		
		//Overwrite the script object
		JSONObject obj = new JSONObject();
		obj.put("k", scriptName);
		obj.put("vars", varNames);
		obj.put("mql", mql);
		obj.put("mem", inMemory);
		
		Record.update(user, "script["+scriptName+"]", obj.toJSONString());
		
	}

	//Delete a script
	public static void delete(String user, String scriptName) throws Exception {
		
		//Generate an error if the script does not exist
		if (!exist(scriptName)) {

			throw new Exception("Sorry, the script "+scriptName+" does not exist.");

		}
		
		scriptName = scriptName.replace("[", "").replace("]", "");
		
		JSONArray groups = GroupManager.getGroupForScript(user, scriptName);
		
		for(int i=0;i<groups.size();i++) {
			
			GroupManager.ungrantScript(user, groups.get(i)+"", scriptName);
			
		}
		
		//Get the script object
		JSONObject bd = Record.getNode("SCRIPT[]");
		
		bd.remove(scriptName);

		Record.update(user, "SCRIPT[]", bd.toJSONString());
		
		Record.remove(user, "script["+scriptName+"]");
		
	}
	
	//Check if a script is granted to a user
	public static boolean isGrantedToUser(String user, String login, String scriptName) throws Exception {
		
		if (login.length()==0) {
			
			throw new Exception("Sorry, the login is required.");
			
		}
		
		if (scriptName.length()==0) {
			
			throw new Exception("Sorry, the script name is required.");
			
		}
		
		if (!UserManager.exist(login)) {
			
			throw new Exception("Sorry, the user <"+login+"> does not exist.");
			
		} else if (!ScriptManager.exist(scriptName)) {
			
			throw new Exception("Sorry, the script <"+scriptName+"> does not exist.");
			
		} else {
			
			boolean result = false;
			BasicData bd = new BasicData();  
			bd.load(GroupManager.showAllGroups());
			
			JSONObject groups = bd.dataNode;
			
			for(Object o:groups.keySet()) {
				
				String groupName = (String) o;
				JSONObject bdGroup = Record.getNode("G["+groupName+"]");
				
				if (((JSONArray) bdGroup.get("users")).contains(login) 
						&& ((JSONArray) bdGroup.get("scripts")).contains(scriptName)) {
					
					result = true;
					break;
					
				}
				
			}
			
			return result;
			
		}
		
	}

}
